# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from unittest import mock, TestCase

import pytest

from tsfpga.module import get_modules
from tsfpga.system_utils import create_file, delete
from tsfpga.vivado.ip_cores import VivadoIpCores
from tsfpga.test.conftest import fixture_tmp_path  # pylint: disable=unused-import


@pytest.mark.usefixtures("fixture_tmp_path")
class TestVivadoIpCores(TestCase):

    tmp_path = None

    def setUp(self):
        self.project_folder = self.tmp_path / "ip_project"
        self.modules_folder = self.tmp_path / "modules"

        self.apa_tcl = create_file(self.modules_folder / "apa" / "ip_cores" / "apa.tcl", "apa")
        self.hest_tcl = create_file(self.modules_folder / "hest" / "ip_cores" / "hest.tcl", "hest")

        modules = get_modules([self.modules_folder])
        self.vivado_ip_cores = VivadoIpCores(modules, self.project_folder, part_name="-")

        # Create initial hash and (empty) compile order file
        self.vivado_ip_cores._save_hash()  # pylint: disable=protected-access
        create_file(self.vivado_ip_cores.compile_order_file)

    @mock.patch("tsfpga.vivado.ip_cores.VivadoProject.create", autospec=True)
    def test_should_not_recreate(self, create):
        assert not self.vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_not_called()

    @mock.patch("tsfpga.vivado.ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_compile_order_file_is_missing(self, create):
        delete(self.vivado_ip_cores.compile_order_file)
        assert self.vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()

    @mock.patch("tsfpga.vivado.ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_hash_file_is_missing(self, create):
        delete(self.vivado_ip_cores._hash_file)  # pylint: disable=protected-access
        assert self.vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()

    @mock.patch("tsfpga.vivado.ip_cores.VivadoProject.create", autospec=True)
    def test_should_not_recreate_if_nothing_is_changed(self, create):
        # This test shows that the pattern used in the upcoming tests:
        #   change something -> get modules -> create new VivadoIpCores object
        # should not result in a recreate unless we actually change something.
        modules = get_modules([self.modules_folder])
        vivado_ip_cores = VivadoIpCores(modules, self.project_folder, part_name="-")

        assert not vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_not_called()

    @mock.patch("tsfpga.vivado.ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_ip_core_file_is_added(self, create):
        create_file(self.modules_folder / "zebra" / "ip_cores" / "zebra.tcl", "zebra")
        modules = get_modules([self.modules_folder])
        vivado_ip_cores = VivadoIpCores(modules, self.project_folder, part_name="-")

        assert vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()

    @mock.patch("tsfpga.vivado.ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_ip_core_file_is_removed(self, create):
        delete(self.hest_tcl)
        modules = get_modules([self.modules_folder])
        vivado_ip_cores = VivadoIpCores(modules, self.project_folder, part_name="-")

        assert vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()

    @mock.patch("tsfpga.vivado.ip_cores.VivadoProject.create", autospec=True)
    def test_should_recreate_if_ip_core_file_is_changed(self, create):
        create_file(self.apa_tcl, "blaha blaha")
        modules = get_modules([self.modules_folder])
        vivado_ip_cores = VivadoIpCores(modules, self.project_folder, part_name="-")

        assert vivado_ip_cores.create_vivado_project_if_needed()
        create.assert_called_once()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path
from shutil import which
import subprocess
import sys

PATH_TO_TSFPGA = Path(__file__).parent.parent.resolve()
sys.path.append(str(PATH_TO_TSFPGA))
PATH_TO_VUNIT = PATH_TO_TSFPGA.parent / "vunit"
sys.path.append(str(PATH_TO_VUNIT))

from vunit import VUnitCLI, VUnit
from vunit.vivado.vivado import create_compile_order_file, add_from_compile_order_file

import tsfpga
import tsfpga.create_vhdl_ls_config
from tsfpga.system_utils import create_directory
from tsfpga.vivado.ip_cores import VivadoIpCores
from tsfpga.vivado.simlib import VivadoSimlib

from examples.tsfpga_example_env import get_tsfpga_modules, TSFPGA_EXAMPLES_TEMP_DIR


def main():
    args = arguments()

    vunit_proj = VUnit.from_args(args=args)
    vunit_proj.add_verification_components()
    vunit_proj.add_random()
    vunit_proj.enable_location_preprocessing()
    vunit_proj.enable_check_preprocessing()

    all_modules = get_tsfpga_modules()
    has_commercial_simulator = vunit_proj.get_simulator_name() != "ghdl"

    if has_commercial_simulator and not args.vivado_skip:
        # Includes modules with IP cores. Can only be used with a commercial simulator.
        sim_modules = all_modules
    else:
        # Only modules that do not contain IP cores
        sim_modules = [module for module in all_modules if len(module.get_ip_core_files()) == 0]

    if args.vivado_skip:
        ip_core_vivado_project_sources_directory = None
    else:
        add_simlib(vunit_proj, args.temp_dir, args.simlib_compile)

        # Generate IP core simulation files. Will be used for the vhdl_ls config,
        # even if they are not added to the simulation project.
        ip_core_compile_order_file, ip_core_vivado_project_sources_directory \
            = generate_ip_core_files(all_modules, args.temp_dir, args.ip_compile)
        if has_commercial_simulator:
            add_from_compile_order_file(vunit_proj, ip_core_compile_order_file)

    create_vhdl_ls_configuration(PATH_TO_TSFPGA,
                                 vunit_proj,
                                 all_modules,
                                 ip_core_vivado_project_sources_directory)

    for module in sim_modules:
        vunit_library = vunit_proj.add_library(module.library_name)
        for hdl_file in module.get_simulation_files():
            if hdl_file.is_vhdl or hdl_file.is_verilog_source:
                vunit_library.add_source_file(hdl_file.path)
            else:
                assert False, f"Can not handle this file: {hdl_file}"
        module.setup_simulations(vunit_proj)

    if vunit_proj.simulator_supports_coverage():
        vunit_proj.set_compile_option("enable_coverage", True)
        vunit_proj.set_sim_option("enable_coverage", True)
        vunit_proj.main(post_run=merge_ghdl_coverage)
    else:
        # Don't enable psl when running with coverage, as the gcc
        # backend seems to have some issues with psl
        vunit_proj.set_compile_option("ghdl.a_flags", ["-fpsl"])

        vunit_proj.main()


def arguments():
    cli = VUnitCLI()
    cli.parser.add_argument("--temp-dir",
                            type=Path,
                            default=TSFPGA_EXAMPLES_TEMP_DIR,
                            help="where to place files needed for simulation flow")
    cli.parser.add_argument("--vivado-skip",
                            action="store_true",
                            help="skip all steps that require Vivado")
    cli.parser.add_argument("--ip-compile",
                            action="store_true",
                            help="force (re)compile of IP cores")
    cli.parser.add_argument("--simlib-compile",
                            action="store_true",
                            help="force (re)compile of Vivado simlib")

    args = cli.parse_args()
    args.output_path = args.temp_dir / "vunit_out"
    return args


def add_simlib(vunit_proj, temp_dir, force_compile):
    vivado_simlib = VivadoSimlib.init(temp_dir, vunit_proj)
    if force_compile or vivado_simlib.compile_is_needed:
        vivado_simlib.compile()
        vivado_simlib.to_archive()
    vivado_simlib.add_to_vunit_project()


def generate_ip_core_files(modules, temp_dir, force_generate):
    vivado_ip_cores = VivadoIpCores(modules, temp_dir, "xc7z020clg400-1")

    if force_generate:
        vivado_ip_cores.create_vivado_project()
        vivado_project_created = True
    else:
        vivado_project_created = vivado_ip_cores.create_vivado_project_if_needed()

    if vivado_project_created:
        # If the IP core Vivado project has been (re)created we need to create
        # a new compile order file
        create_compile_order_file(vivado_ip_cores.vivado_project_file,
                                  vivado_ip_cores.compile_order_file)

    return vivado_ip_cores.compile_order_file, vivado_ip_cores.vivado_project_sources_directory


def create_vhdl_ls_configuration(output_path, vunit_proj, modules, ip_core_vivado_project_sources_directory):
    """
    Create config for vhdl_ls. Granted this might no be the "correct" place for this functionality.
    But since the call is somewhat quick (~10 ms), and simulate.py is run "often" it seems an
    appropriate place in order to always have an up-to-date vhdl_ls config.
    """
    vivado_location = None if which("vivado") is None else Path(which("vivado"))
    tsfpga.create_vhdl_ls_config.create_configuration(
        output_path=output_path,
        modules=modules,
        vunit_proj=vunit_proj,
        vivado_location=vivado_location,
        ip_core_vivado_project_sources_directory=ip_core_vivado_project_sources_directory)


def merge_ghdl_coverage(results):
    if not results.get_report().tests:
        print("No test results to merge coverage on")
        return

    temp_dir = results.get_report().output_path.parent

    merged_coverage_output = temp_dir / "vunit_coverage_database"
    results.merge_coverage(merged_coverage_output)

    report_html_path = temp_dir / "vhdl_coverage_html"
    create_directory(report_html_path, empty=True)
    gcovr_cmd = ["gcovr",
                 "--xml",
                 temp_dir / "vhdl_coverage.xml",
                 "--xml-pretty",
                 "--html",
                 "--html-details",
                 report_html_path / "index.html",
                 "--html-title",
                 "tsfpga VHDL coverage",
                 "--exclude-unreachable-branches",
                 "--exclude-throw-branches",
                 merged_coverage_output,
                 ]
    subprocess.call(gcovr_cmd)

    print(f"Coverage HTML report saved in {report_html_path}")


if __name__ == "__main__":
    main()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.module import BaseModule
from tsfpga.vivado.project import VivadoNetlistProject
from tsfpga.vivado.size_checker import *  # pylint: disable=wildcard-import,unused-wildcard-import
from examples.tsfpga_example_env import get_tsfpga_modules


class Module(BaseModule):

    def setup_simulations(self, vunit_proj, **kwargs):
        for test in vunit_proj.library(self.library_name).test_bench("tb_afifo").get_tests():
            for read_clock_is_faster in [True, False]:
                original_generics = dict(read_clock_is_faster=read_clock_is_faster)

                for generics in self.generate_common_fifo_test_generics(test.name, original_generics):
                    self.add_config(test, generics=generics)

        for test in vunit_proj.library(self.library_name).test_bench("tb_fifo").get_tests():
            for generics in self.generate_common_fifo_test_generics(test.name):
                self.add_config(test, generics=generics)

    @staticmethod
    def generate_common_fifo_test_generics(test_name, original_generics=None):
        generics = original_generics if original_generics is not None else dict()

        if "init_state" in test_name or "almost" in test_name:
            # Note that
            #   almost_full_level = depth, or
            #   almost_empty_level = 0
            # result in alternative ways of calculating almost full/empty.
            depth = 32

            for almost_full_level, almost_empty_level in [(depth, depth // 2), (depth // 2, 0)]:
                generics.update(depth=depth,
                                almost_full_level=almost_full_level,
                                almost_empty_level=almost_empty_level)

                yield generics

        else:
            if "write_faster_than_read" in test_name:
                generics.update(read_stall_probability_percent=90)
                generics.update(enable_last=True)
            if "read_faster_than_write" in test_name:
                generics.update(write_stall_probability_percent=90)
            if "packet_mode" in test_name:
                generics.update(enable_packet_mode=True, enable_last=True)

            for depth in [16, 512]:
                generics.update(depth=depth)

                yield generics

    def setup_formal(self, formal_proj, **kwargs):
        depth = 4
        for (almost_full_level, almost_empty_level) in [(depth - 1, 0), (depth, 1)]:
            generics = dict(
                width=3,
                depth=depth,
                almost_full_level=almost_full_level,
                almost_empty_level=almost_empty_level)
            formal_proj.add_config(top="fifo", generics=generics)

    def get_build_projects(self):
        projects = []
        all_modules = get_tsfpga_modules()
        part = "xc7z020clg400-1"

        # Use a wrapper as top level, which only routes the "barebone" ports, resulting in
        # a minimal FIFO.
        generics = dict(width=32, depth=1024)
        projects.append(VivadoNetlistProject(
            name=self.test_case_name("fifo_minimal", generics),
            modules=all_modules,
            part=part,
            top="fifo_netlist_build_wrapper",
            generics=generics,
            result_size_checkers=[
                TotalLuts(EqualTo(16)),
                LogicLuts(EqualTo(16)),
                Ffs(EqualTo(24)),
                Ramb36(EqualTo(1)),
                Ramb18(EqualTo(0)),
            ]
        ))

        # A FIFO with level counter port and non-default almost_full_level, which
        # increases resource utilization.
        generics = dict(width=32, depth=1024, almost_full_level=800)
        projects.append(VivadoNetlistProject(
            name=self.test_case_name("fifo_regular", generics),
            modules=all_modules,
            part=part,
            top="fifo",
            generics=generics,
            result_size_checkers=[
                TotalLuts(EqualTo(29)),
                LogicLuts(EqualTo(29)),
                Ffs(EqualTo(35)),
                Ramb36(EqualTo(1)),
                Ramb18(EqualTo(0)),
            ]
        ))

        return projects



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import pytest

from tsfpga.system_utils import create_directory
from tsfpga.vivado.project import VivadoProject


def test_create_should_raise_exception_if_project_path_already_exists(tmp_path):
    create_directory(tmp_path / "projects" / "name")
    proj = VivadoProject(name="name", modules=[], part="part")
    with pytest.raises(ValueError) as exception_info:
        proj.create(tmp_path / "projects")
    assert str(exception_info.value).startswith("Folder already exists")


def test_build_should_raise_exception_if_project_does_not_exists(tmp_path):
    create_directory(tmp_path / "projects")
    proj = VivadoProject(name="name", modules=[], part="part")
    with pytest.raises(ValueError) as exception_info:
        proj.build(tmp_path / "projects", synth_only=True)
    assert str(exception_info.value).startswith("Project file does not exist")


def test_build_with_impl_run_should_raise_exception_if_no_output_path_is_given():
    proj = VivadoProject(name="name", modules=[], part="part")
    with pytest.raises(ValueError) as exception_info:
        proj.build("None")
    assert str(exception_info.value).startswith("Must specify output_path")



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import subprocess

import pytest

from tsfpga.system_utils import run_command


def test_run_command_called_with_nonexisting_binary_should_raise_exception():
    cmd = ["/apa/hest/zebra.exe", "foobar"]
    with pytest.raises(FileNotFoundError):
        run_command(cmd)


def test_run_command_with_non_zero_return_code_should_raise_exception():
    cmd = ["ls", "/apa/hest/zebra"]
    with pytest.raises(subprocess.CalledProcessError):
        run_command(cmd)


def test_run_command_called_with_non_list_should_raise_exception():
    cmd = ["ls", "-la"]
    run_command(cmd)

    cmd = "ls -la"
    with pytest.raises(ValueError) as exception_info:
        run_command(cmd)
    assert str(exception_info.value).startswith("Must be called with a list")



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import hashlib
from pathlib import Path

from tsfpga.system_utils import create_file
from .common import get_vivado_path, run_vivado_tcl, to_tcl_path
from .simlib_common import VivadoSimlibCommon


class VivadoSimlibCommercial(VivadoSimlibCommon):

    """
    Handle Vivado simlib with a commercial simulator.
    """

    _libraries = ["unisim", "secureip", "unimacro", "unifast", "xpm"]

    _tcl = "set_param general.maxthreads 4\n" \
        "compile_simlib -simulator {simulator_name} -simulator_exec_path {{{simulator_folder}}} " \
        "-family all -language all -library all -no_ip_compile -dir {{{output_path}}} -force"

    def __init__(self, output_path, vunit_proj, simulator_interface, vivado_path):
        """
        Args:
            output_path (`pathlib.Path`): The compiled simlib will be placed here.
            vunit_proj: The VUnit project that is used to run simulation.
            simulator_interface: A VUnit SimulatorInterface class.
            vivado_path (`pathlib.Path`): Path to Vivado executable.
        """
        self._vunit_proj = vunit_proj
        self._vivado_path = get_vivado_path(vivado_path)

        # Vivado uses a different name for Riviera-PRO
        self._simulator_name = "riviera" if simulator_interface.name == "rivierapro" \
            else simulator_interface.name
        self._simulator_folder = Path(simulator_interface.find_prefix())

        self._output_path = output_path / self._get_version_tag()

    def _compile(self):
        tcl_file = self._output_path / "compile_simlib.tcl"
        tcl = self._tcl.format(simulator_name=self._simulator_name,
                               simulator_folder=to_tcl_path(self._simulator_folder),
                               output_path=to_tcl_path(self._output_path))
        create_file(tcl_file, tcl)
        run_vivado_tcl(self._vivado_path, tcl_file)

    def _get_version_tag(self):
        """
        Add a hash of the TCL script to version tag.
        """
        tag = super()._get_version_tag()
        tag += "." + self._get_tcl_hash()
        return tag

    def _get_simulator_tag(self):
        """
        Return e.g. modelsim_modeltech_pe_10_6c or riviera_riviera_pro_2018_10_x64.
        """
        simulator_version = self._simulator_folder.parent.name
        return self._format_version(f"{self._simulator_name}_{simulator_version}")

    def _get_tcl_hash(self):
        """
        Return a (partial) hash of the TCL script that is used to compile simlib.
        """
        return hashlib.md5(self._tcl.encode()).hexdigest()[0:8]

    def _add_to_vunit_project(self):
        """
        Add the compiled simlib to your VUnit project.
        """
        for library in self._libraries:
            library_path = self._output_path / library
            assert library_path.exists(), library_path
            self._vunit_proj.add_external_library(library, library_path)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import re


class UtilizationParser():

    """
    Used for parsing the utilization report generated by Vivado
    """

    @staticmethod
    def get_size(report):
        """
        Takes a hierarchical utilization report as a string and returns the top level size
        for the specified run.

        Args:
            report (string): A string containing the entire Vivado hierarchical utilization report.
        """
        result = dict()
        lines = report.split("\n")
        for idx, line in enumerate(lines):
            # Find the table line that is the top level
            if re.search(r"\(top\)", line):
                # Parse the report, remove uninteresting fields and create dictionary
                # Note that "|" is the column separator. Heading titles for the data is two lines above
                # the row for the top level.
                headers = [column_data.strip() for column_data in lines[idx - 2].split("|")]
                numbers = [column_data.strip() for column_data in line.split("|")]
                # The first columns contain entity name, etc. We only want the numbers
                headers = headers[3:-1]
                numbers = numbers[3:-1]
                # Convert numbers from string to integers
                numbers = [int(number) for number in numbers]
                result = dict(zip(headers, numbers))
                break

        return result


class LessThan:

    """
    Limit to be used with a checker to see that a resource number is less than the
    specified value.
    """

    def __init__(self, value):
        """
        Args:
            value (int): The result size value shall be less than this.
        """
        self.value = value

    def check(self, result_value):
        return result_value < self.value

    def __str__(self):
        return f"< {self.value}"


class EqualTo:

    """
    Limit to be used with a checker to see that a resource number is equal to the
    specified value.
    """

    def __init__(self, value):
        """
        Args:
            value (int): The result size value shall be equal to this.
        """
        self.value = value

    def check(self, result_value):
        return result_value == self.value

    def __str__(self):
        return str(self.value)


class SizeChecker:

    """
    Check a build result size value against a limit.

    Overload and set the correct ``resource_name``, according to the names
    in the vendor utilization report.
    """

    resource_name = ""

    def __init__(self, limit):
        """
        Args:
            limit: The limit that the specified resource shall be checked against. Should
                be e.g. a :class:`LessThan` object.
        """
        self.limit = limit

    def check(self, result_size):
        """
        Args:
            result_size (dict): A size dictionary as found in the :class:`.BuildResult`
                object.
        """
        result_value = result_size[self.resource_name]
        if self.limit.check(result_value):
            message = f"Result size check passed for {self.resource_name}: "\
                f"{result_value} ({self.limit})"
            print(message)
            return True

        message = f"Result size check failed for {self.resource_name}. "\
            f"Got {result_value}, expected {self.limit}."
        print(message)
        return False


class TotalLuts(SizeChecker):

    resource_name = "Total LUTs"


class LogicLuts(SizeChecker):

    resource_name = "Logic LUTs"


class Dsp48Blocks(SizeChecker):

    resource_name = "DSP48 Blocks"


class Ffs(SizeChecker):

    resource_name = "FFs"


class Ramb36(SizeChecker):

    resource_name = "RAMB36"


class Ramb18(SizeChecker):

    resource_name = "RAMB18"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.module import BaseModule


class Module(BaseModule):

    def setup_simulations(self, vunit_proj, **kwargs):
        tb = vunit_proj.library(self.library_name).test_bench("tb_axi_pkg")
        for data_width in [32, 64]:
            for id_width in [0, 5]:
                for addr_width in [32, 40]:
                    generics = dict(data_width=data_width, id_width=id_width, addr_width=addr_width)
                    self.add_config(tb, generics=generics)

        tb = vunit_proj.library(self.library_name).test_bench("tb_axil_pkg")
        for data_width in [32, 64]:
            for addr_width in [32, 40]:
                generics = dict(data_width=data_width, addr_width=addr_width)
                self.add_config(tb, generics=generics)

        for tb_name in ["tb_axi_to_axil", "tb_axi_to_axil_bus_error"]:
            tb = vunit_proj.library(self.library_name).test_bench(tb_name)
            for data_width in [32, 64]:
                name = "data_width_%i" % data_width
                tb.add_config(name=name, generics=dict(data_width=data_width))

        tb = vunit_proj.library(self.library_name).test_bench("tb_axil_cdc")
        tb.add_config(name="master_clk_fast", generics=dict(master_clk_fast=True))
        tb.add_config(name="slave_clk_fast", generics=dict(slave_clk_fast=True))
        tb.add_config(name="same_clocks")

        tb = vunit_proj.library(self.library_name).test_bench("tb_axi_fifo")
        tb.add_config(name="passthrough", generics=dict(depth=0))
        tb.add_config(name="synchronous", generics=dict(depth=16, asynchronous=False))
        tb.add_config(name="asynchronous", generics=dict(depth=16, asynchronous=True))

        tb = vunit_proj.library(self.library_name).test_bench("tb_axi_interconnect")
        tb.add_config(name="axi_lite", generics=dict(test_axi_lite=True))
        tb.add_config(name="axi", generics=dict(test_axi_lite=False))

        tb = vunit_proj.library(self.library_name).test_bench("tb_axil_mux")
        tb.test("read_from_non_existent_slave_base_adress").set_generic("use_axil_bfm", False)
        tb.test("write_to_non_existent_slave_base_adress").set_generic("use_axil_bfm", False)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from os.path import relpath
from pathlib import Path
import sys

from setuptools import setup, find_packages

REPO_ROOT = Path(__file__).parent
sys.path.insert(0, str(REPO_ROOT))
import tsfpga
from tsfpga.about import get_slogan
from tsfpga.system_utils import read_file


README_RST = REPO_ROOT / "readme.rst"
REQUIREMENTS_TXT = REPO_ROOT / "requirements.txt"
REQUIREMENTS_DEVELOP_TXT = REPO_ROOT / "requirements_develop.txt"


def main():
    setup(
        name="tsfpga",
        version=tsfpga.__version__,
        description=get_slogan(),
        long_description=read_file(README_RST),
        long_description_content_type="text/x-rst",
        license="BSD 3-Clause License",
        author="Lukas Vik",
        author_email="2767848-LukasVik@users.noreply.gitlab.com",
        url="https://tsfpga.com",
        python_requires=">=3.6",
        install_requires=read_requirements_file(REQUIREMENTS_TXT),
        extras_require=dict(dev=read_requirements_file(REQUIREMENTS_DEVELOP_TXT)),
        packages=find_packages(),
        package_data={"tsfpga": get_package_data()},
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "License :: OSI Approved :: BSD License",
            "Natural Language :: English",
            "Intended Audience :: Developers",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX :: Linux",
            "Topic :: Software Development :: Testing",
            "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        ],
        zip_safe=False
    )


def read_requirements_file(path):
    requirements = []
    with path.open() as file_handle:
        # Requirements file contains one package name per line
        for line_data in file_handle.readlines():
            if line_data:
                requirements.append(line_data.strip())

    return requirements


def get_package_data():
    """
    Additional files that shall be included in the release, apart from the python packages
    """
    package_data = [
        README_RST,
        REQUIREMENTS_TXT,
        REQUIREMENTS_DEVELOP_TXT,
        tsfpga.TSFPGA_PATH / "test" / "lint" / "pylintrc",
        tsfpga.TSFPGA_PATH / "test" / "lint" / "pycodestylerc",
    ]
    package_data += find_package_files(tsfpga.TSFPGA_TCL)
    package_data += find_package_files(tsfpga.TSFPGA_MODULES)

    # Specify path relative to the tsfpga python package folder
    tsfpga_root = Path(REPO_ROOT, "tsfpga")
    package_data = [path_relative_to_str(file, tsfpga_root) for file in package_data]
    return package_data


def path_relative_to_str(path, other):
    """
    Note Path.relative_to() does not support the use case where e.g. readme.md should get
    relative path "../readme.md". Hence we have to use os.path.
    """
    return relpath(str(path), str(other))


def find_package_files(directory):
    """
    Find files to include in the package, ignoring temporary files.
    """
    files = []
    for file in directory.glob("**/*"):
        if not file.suffix.endswith("pyc"):
            files.append(file)
    return files


if __name__ == "__main__":
    main()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.module import BaseModule


class Module(BaseModule):

    def setup_simulations(self, vunit_proj, **kwargs):
        tb = vunit_proj.library(self.library_name).test_bench("tb_width_conversion")

        for input_width in [8, 16, 32]:
            for output_width in [8, 16, 32]:
                if input_width == output_width:
                    continue
                test = tb.get_tests("test_data")[0]
                name = "input_%i.output_%s" % (input_width, output_width)
                test.add_config(name=name, generics=dict(input_width=input_width, output_width=output_width))

        test = tb.get_tests("test_full_throughput")[0]
        test.add_config(name="input_16.output_8", generics=dict(input_width=16, output_width=8, data_jitter=False))
        test.add_config(name="input_8.output_16", generics=dict(input_width=8, output_width=16, data_jitter=False))

        test = vunit_proj.library(self.library_name).test_bench("tb_handshake_pipeline").get_tests("test_full_throughput")[0]
        test.set_generic("data_jitter", False)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.system_utils import create_file
from .common import to_tcl_path


class VivadoTcl:
    """
    Class with methods for translating a set of sources into Vivado TCL
    """

    def __init__(self, name,):
        self.name = name

    # pylint: disable=too-many-arguments
    def create(self,
               project_folder,
               modules,
               part,
               top,
               run_index,
               generics=None,
               constraints=None,
               tcl_sources=None,
               build_step_hooks=None,
               ip_cache_path=None,
               disable_io_buffers=True):
        tcl = f"create_project {self.name} {to_tcl_path(project_folder)} -part {part}\n"
        tcl += "set_property target_language VHDL [current_project]\n"
        if ip_cache_path is not None:
            tcl += f"config_ip_cache -use_cache_location {to_tcl_path(ip_cache_path)}\n"
        tcl += "\n"
        tcl += self._add_modules(modules)
        tcl += "\n"
        tcl += self._add_generics(generics)
        tcl += "\n"
        tcl += self._add_constraints(self._iterate_constraints(modules, constraints))
        tcl += "\n"
        tcl += self._add_tcl_sources(tcl_sources)
        tcl += "\n"
        tcl += self._add_build_step_hooks(build_step_hooks, project_folder)
        tcl += "\n"
        tcl += self._add_project_settings()
        tcl += "\n"
        tcl += f"current_run [get_runs synth_{run_index}]\n"
        tcl += "\n"
        tcl += f"set_property top {top} [current_fileset]\n"
        tcl += "reorder_files -auto -disable_unused\n"
        tcl += "\n"
        if disable_io_buffers:
            tcl += f"set_property -name {{STEPS.SYNTH_DESIGN.ARGS.MORE OPTIONS}} -value -no_iobuf -objects [get_runs synth_{run_index}]"
        tcl += "\n"
        tcl += "exit\n"
        return tcl

    def _add_modules(self, modules):
        tcl = ""
        for module in modules:
            vhdl_files = []
            verilog_source_files = []
            for hdl_file in module.get_synthesis_files():
                if hdl_file.is_vhdl:
                    vhdl_files.append(to_tcl_path(hdl_file.path))
                elif hdl_file.is_verilog_source:
                    verilog_source_files.append(to_tcl_path(hdl_file.path))
                else:
                    raise NotImplementedError(f"Can not handle file: {hdl_file}")
                    # Verilog headers do not need to be handled at all if the
                    # source file that uses them is in the same directory. If
                    # it is not, the path needs to be added to include_dirs with
                    # a tcl command like:
                    #   set_property include_dirs {/some/path /some/other/path} [current_fileset]
                    # See https://www.xilinx.com/support/answers/54006.html

                    # Encrypted source files (verilog (.vp?), VHDL) I do not know how
                    # to handle, since I have no use case for it at the moment.

            if vhdl_files:
                tcl += "read_vhdl -library %s -vhdl2008 {%s}\n" % \
                    (module.library_name, " ".join(vhdl_files))
            if verilog_source_files:
                tcl += "read_verilog {%s}\n" % " ".join(verilog_source_files)

            tcl += self._add_tcl_sources(module.get_ip_core_files())
        return tcl

    @staticmethod
    def _add_tcl_sources(tcl_sources):
        if tcl_sources is None:
            return ""

        tcl = ""
        for tcl_source_file in tcl_sources:
            tcl += "source -notrace %s\n" % to_tcl_path(tcl_source_file)
        return tcl

    def _add_build_step_hooks(self, build_step_hooks, project_folder):
        if build_step_hooks is None:
            return ""

        # There can be many hooks for the same step. Reorganize them into a dict, according
        # to the format step_name: [list of hooks]
        hook_steps = dict()
        for build_step_hook in build_step_hooks:
            if build_step_hook.hook_step in hook_steps:
                hook_steps[build_step_hook.hook_step].append(build_step_hook)
            else:
                hook_steps[build_step_hook.hook_step] = [build_step_hook]

        tcl = ""
        for step, hooks in hook_steps.items():
            # Vivado will only accept one TCL script as hook for each step. So if we want
            # to add more we have to create a new TCL file, that sources the other files,
            # and add that as the hook to Vivado.
            if len(hooks) == 1:
                tcl_file = hooks[0].tcl_file
            else:
                tcl_file = project_folder / ("hook_" + step.replace(".", "_") + ".tcl")
                source_hooks_tcl = "".join([f"source {to_tcl_path(hook.tcl_file)}\n" for hook in hooks])
                create_file(tcl_file, source_hooks_tcl)

            # Add to fileset to enable archive and other project based functionality
            tcl += f"add_files -fileset utils_1 -norecurse {to_tcl_path(tcl_file)}\n"

            # Build step hook can only be applied to a run (e.g. impl_1), not on a project basis
            run_wildcard = "synth_*" if hooks[0].step_is_synth else "impl_*"
            tcl_block = f"set_property {step} {to_tcl_path(tcl_file)} ${{run}}"
            tcl += self._tcl_for_each_run(run_wildcard, tcl_block)

        return tcl

    def _add_project_settings(self):
        tcl = ""
        # Default value for when opening project in GUI.
        # Will be overwritten if using build() function.
        tcl += "set_param general.maxThreads 7\n"
        # Enable VHDL assert statements to be evaluated. A severity level of failure will
        # stop the synthesis and produce an error.
        tcl_block = "set_property STEPS.SYNTH_DESIGN.ARGS.ASSERT true ${run}"
        tcl += self._tcl_for_each_run("synth_*", tcl_block)
        # Enable binary bitstream as well
        tcl_block = "set_property STEPS.WRITE_BITSTREAM.ARGS.BIN_FILE true ${run}"
        tcl += self._tcl_for_each_run("impl_*", tcl_block)
        return tcl

    @staticmethod
    def _tcl_for_each_run(run_wildcard, tcl_block):
        """
        Apply TCL block for each defined run. Use ${run} for run variable in TCL.
        """
        tcl = ""
        tcl += f"foreach run [get_runs {run_wildcard}] {{\n"
        tcl += tcl_block + "\n"
        tcl += "}\n"
        return tcl

    @staticmethod
    def _add_generics(generics):
        """
        Generics are set according to this weird format: https://www.xilinx.com/support/answers/52217.html
        """
        if generics is None:
            return ""

        generic_list = []
        for name, value in generics.items():
            if isinstance(value, bool):
                generic_list.append("%s=%s" % (name, ("1'b1" if value else "1'b0")))
            else:
                generic_list.append("%s=%s" % (name, value))
        return "set_property generic {%s} [current_fileset]\n" % " ".join(generic_list)

    @staticmethod
    def _iterate_constraints(modules, constraints):
        for module in modules:
            for constraint in module.get_scoped_constraints():
                yield constraint

        if constraints is not None:
            for constraint in constraints:
                yield constraint

    @staticmethod
    def _add_constraints(constraints):
        tcl = ""
        for constraint in constraints:
            file = to_tcl_path(constraint.file)
            ref_flags = "" if constraint.ref is None else (f"-ref {constraint.ref} ")
            managed_flags = "" if file.endswith("xdc") else "-unmanaged "

            tcl += f"read_xdc {ref_flags}{managed_flags}{file}\n"
            tcl += f"set_property PROCESSING_ORDER {constraint.processing_order} [get_files {file}]\n"

            if constraint.used_in == "impl":
                tcl += f"set_property used_in_synthesis false [get_files {file}]\n"
            elif constraint.used_in == "synth":
                tcl += f"set_property used_in_implementation false [get_files {file}]\n"

        return tcl

    def build(
            self,
            project_file,
            output_path,
            num_threads,
            run_index,
            generics=None,
            synth_only=False,
            analyze_clock_interaction=True):
        # Max value in Vivado 2017.4. set_param will give an error if higher number.
        num_threads = min(num_threads, 8)

        synth_run = f"synth_{run_index}"
        impl_run = f"impl_{run_index}"

        tcl = f"open_project {to_tcl_path(project_file)}\n"
        tcl += f"set_param general.maxThreads {num_threads}\n"
        tcl += "\n"
        tcl += self._add_generics(generics)
        tcl += "\n"
        tcl += self._synthesis(synth_run, num_threads, analyze_clock_interaction)
        tcl += "\n"
        if not synth_only:
            tcl += self._run(impl_run, num_threads, to_step="write_bitstream")
            tcl += "\n"
            tcl += self._write_hw_platform(output_path)
            tcl += "\n"
        tcl += "exit\n"
        return tcl

    def _synthesis(self, run, num_threads, analyze_clock_interaction):
        tcl = self._run(run, num_threads)
        if analyze_clock_interaction:
            tcl += "\n"
            tcl += f"open_run {run}\n"
            tcl += f"set run_directory [get_property DIRECTORY [get_runs {run}]]\n"
            tcl += "\n"
            tcl += r"if {[regexp {\(unsafe\)} [report_clock_interaction -delay_type min_max -return_string]]} "
            tcl += "{\n"
            tcl += f"  puts \"ERROR: Unhandled clock crossing in {run} run. See reports in ${{run_directory}}\"\n"
            tcl += "\n"
            tcl += "  set output_file [file join ${run_directory} \"clock_interaction.rpt\"]\n"
            tcl += "  report_clock_interaction -delay_type min_max -file ${output_file}\n"
            tcl += "\n"
            tcl += "  set output_file [file join ${run_directory} \"timing_summary.rpt\"]\n"
            tcl += "  report_timing_summary -file ${output_file}\n"
            tcl += "\n"
            tcl += "  exit 1\n"
            tcl += "}\n"
        return tcl

    @staticmethod
    def _run(run, num_threads, to_step=None):
        to_step = "" if to_step is None else " -to_step " + to_step

        tcl = f"reset_run {run}\n"
        tcl += f"launch_runs {run} -jobs {num_threads}{to_step}\n"
        tcl += "wait_on_run %s\n" % run
        tcl += "\n"
        tcl += "if {[get_property PROGRESS [get_runs %s]] != \"100%%\"} {\n" % run
        tcl += f"  puts \"ERROR: Run {run} failed.\"\n"
        tcl += "  exit 1\n"
        tcl += "}\n"
        return tcl

    def _write_hw_platform(self, output_path):
        xsa_file = to_tcl_path(output_path / (self.name + ".xsa"))
        tcl = f"write_hw_platform -fixed -force {xsa_file}\n"
        return tcl



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import subprocess
import sys
import unittest

from pathlib import Path
import pytest

import tsfpga
from tsfpga.git_utils import find_git_files
from tsfpga.system_utils import create_file


THIS_DIR = Path(__file__).parent


def run_pylint(files):
    config = THIS_DIR / "pylintrc"
    command = [sys.executable, "-m", "pylint", f"--rcfile={config}"] + files

    subprocess.check_call(command)


def test_pylint():
    # Exclude doc folder, since conf.py used by sphinx does not conform
    files = list(find_git_files(file_endings_include="py", exclude_directories=[tsfpga.TSFPGA_DOC]))
    run_pylint(files)


def run_pycodestyle(files):
    config = THIS_DIR / "pycodestylerc"
    command = [sys.executable, "-m", "pycodestyle", f"--config={config}"] + files

    subprocess.check_call(command)


def test_pycodestyle():
    files = list(find_git_files(file_endings_include="py"))
    run_pycodestyle(files)


@pytest.mark.usefixtures("fixture_tmp_path")
class TestPythonLintFunctions(unittest.TestCase):

    tmp_path = None

    def setUp(self):
        self.file = self.tmp_path / "dummy_python_file.py"
        ugly_code = "aa  =\ndef bb:\ncc  = 3"
        create_file(self.file, ugly_code)

    def test_pylint_should_raise_exception_if_there_are_ugly_files(self):
        with pytest.raises(subprocess.CalledProcessError):
            run_pylint([self.file])

    def test_pycodestyle_should_raise_exception_if_there_are_ugly_files(self):
        with pytest.raises(subprocess.CalledProcessError):
            run_pycodestyle([self.file])



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.markdown_to_html_translator import MarkdownToHtmlTranslator
from tsfpga.register_types import Register, REGISTER_MODES


class RegisterHtmlGenerator:

    def __init__(self, module_name, generated_info):
        self.module_name = module_name
        self.generated_info = generated_info
        self._markdown_to_html = MarkdownToHtmlTranslator()

    @staticmethod
    def _comment(comment):
        return f"<!-- {comment} -->\n"

    def _file_header(self):
        return self._comment(self.generated_info)

    @staticmethod
    def _to_readable_address(address):
        num_nibbles_needed = 4
        formatting_string = "0x{:0%iX}" % num_nibbles_needed
        return formatting_string.format(address)

    def _annotate_register(self, register, register_array_index=None, array_index_increment=None):
        if register_array_index is None:
            address_readable = self._to_readable_address(register.address)
        else:
            register_address = self._to_readable_address(4 * register_array_index)
            address_increment = self._to_readable_address(4 * array_index_increment)
            address_readable = f"{register_address} + i &times; {address_increment}"
        description = self._markdown_to_html.translate(register.description)
        html = f"""
  <tr>
    <td><strong>{register.name}</strong></td>
    <td>{address_readable}</td>
    <td>{REGISTER_MODES[register.mode].mode_readable}</td>
    <td>{register.default_value}</td>
    <td>{description}</td>
  </tr>"""

        return html

    def _annotate_bit(self, bit):
        description = self._markdown_to_html.translate(bit.description)
        html = f"""
  <tr>
    <td>&nbsp;&nbsp;<em>{bit.name}</em></td>
    <td>{bit.index}</td>
    <td></td>
    <td></td>
    <td>{description}</td>
  </tr>"""

        return html

    def _get_table(self, register_objects):
        html = """
<table>
<thead>
  <tr>
    <th>Name</th>
    <th>Address</th>
    <th>Mode</th>
    <th>Default value</th>
    <th>Description</th>
  </tr>
</thead>
<tbody>"""

        for register_object in register_objects:
            if isinstance(register_object, Register):
                html += self._annotate_register(register_object)
                for bit in register_object.bits:
                    html += self._annotate_bit(bit)
            else:
                html += f"""
  <tr>
    <td colspan="5" class="array_header">Register array <strong>{register_object.name}</strong>, repeated {register_object.length} times</td>
  </tr>"""
                array_index_increment = len(register_object.registers)
                for register in register_object.registers:
                    register_index = register_object.base_index + register.index
                    html += self._annotate_register(register, register_index, array_index_increment)
                html += f"""
  <tr>
    <td colspan="5" class="array_header">End register array <strong>{register_object.name}</strong></td>
  </tr>"""

        html += """
</tbody>
</table>"""

        return html

    def get_table(self, register_objects):
        html = self._file_header()
        html += self._get_table(register_objects)
        return html

    @staticmethod
    def _get_mode_descriptions():
        html = """
<table>
<thead>
  <tr>
    <th>Mode</th>
    <th>Description</th>
  </tr>
</thead>
<tbody>"""

        for mode in REGISTER_MODES.values():
            html += f"""
<tr>
  <td>{mode.mode_readable}</td>
  <td>{mode.description}</td>
</tr>
"""
        html += """
</tbody>
</table>"""
        return html

    def get_page(self, register_objects, table_style=None, font_style=None, extra_style=""):
        title = f"Documentation of {self.module_name} registers"

        if font_style is None:
            font_style = """
html * {
  font-family: "Trebuchet MS", Arial, Helvetica, sans-serif;
}"""

        if table_style is None:
            table_style = """
table {
  border-collapse: collapse;
}
td, th {
  border: 1px solid #ddd;
  padding: 8px;
}
td.array_header {
  background-color: #4cacaf;
  color: white;
}
tr:nth-child(even) {
  background-color: #f2f2f2;
}
tr:hover {
  background-color: #ddd;
}
th {
  padding-top: 12px;
  padding-bottom: 12px;
  text-align: left;
  background-color: #4CAF50;
  color: white;
}"""

        html = f"""
{self._file_header()}

<!DOCTYPE html>
<html>
<head>
  <title>{title}</title>
  <style>
{font_style}
{table_style}
{extra_style}
  </style>
</head>
<body>
  <h1>{title}</h1>
  <p>This document is a specification of the PS interface of the {self.module_name} module.</p>
  <h2>Register modes</h2>
  <p>The following register modes are available.</p>
{self._get_mode_descriptions()}
  <h2>Register list</h2>
  <p>The following registers make up the register map for the {self.module_name} module.</p>
{self._get_table(register_objects)}
<p>{self.generated_info}</p>
</body>
</html>"""

        return html



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from pathlib import Path
from unittest import TestCase
from unittest.mock import MagicMock, patch

import pytest

from tsfpga.vivado.simlib import VivadoSimlib
from tsfpga.test.conftest import fixture_tmp_path  # pylint: disable=unused-import


@pytest.mark.usefixtures("fixture_tmp_path")
class TestVivadoSimlibCommercial(TestCase):

    tmp_path = None

    def setUp(self):
        self.output_path = self.tmp_path / "simlib"

        self.simulator_prefix = "/opt/Aldec/Riviera-PRO-2018.10-x64/bin"
        self.vivado_path = Path("/tools/xilinx/Vivado/2019.2/bin/vivado")

        self.vivado_simlib = self.get_vivado_simlib(self.simulator_prefix, self.vivado_path)

    def get_vivado_simlib(self, simulator_prefix, vivado_path):
        simulator_class = MagicMock()
        simulator_class.name = "rivierapro"
        simulator_class.find_prefix.return_value = simulator_prefix

        vunit_proj = MagicMock()
        vunit_proj._simulator_class = simulator_class  # pylint: disable=protected-access

        return VivadoSimlib.init(self.output_path, vunit_proj, vivado_path)

    def test_should_not_recompile(self):
        self.assert_should_compile(self.vivado_simlib)
        self.assert_should_not_compile(self.vivado_simlib)

    def test_new_simulator_version_should_cause_recompile(self):
        self.assert_should_compile(self.vivado_simlib)
        self.assert_should_not_compile(self.vivado_simlib)

        vivado_simlib = self.get_vivado_simlib("/opt/Aldec/Riviera-PRO-1975.01-x64/bin",
                                               self.vivado_path)
        self.assert_should_compile(vivado_simlib)
        self.assert_should_not_compile(vivado_simlib)

    def test_new_vivado_version_should_cause_recompile(self):
        self.assert_should_compile(self.vivado_simlib)
        self.assert_should_not_compile(self.vivado_simlib)

        vivado_simlib = self.get_vivado_simlib(self.simulator_prefix,
                                               Path("/tools/xilinx/Vivado/1337.2/bin/vivado"))
        self.assert_should_compile(vivado_simlib)
        self.assert_should_not_compile(vivado_simlib)

    @staticmethod
    def assert_should_compile(vivado_simlib):
        assert vivado_simlib.compile_is_needed
        with patch("tsfpga.vivado.simlib_commercial.run_vivado_tcl", autospec=True) as mock:
            vivado_simlib.compile_if_needed()
            mock.assert_called_once()

    @staticmethod
    def assert_should_not_compile(vivado_simlib):
        assert not vivado_simlib.compile_is_needed
        with patch("tsfpga.vivado.simlib_commercial.run_vivado_tcl", autospec=True) as mock:
            vivado_simlib.compile_if_needed()
            mock.assert_not_called()


@pytest.mark.usefixtures("fixture_tmp_path")
class TestVivadoSimlibGhdl(TestCase):

    """
    Test a subset of what is tested for commercial, since most of the code is inherited
    from the common class.
    """

    tmp_path = None

    def setUp(self):
        self.output_path = self.tmp_path / "simlib"

        self.vivado_simlib = self.get_vivado_simlib()

    def get_vivado_simlib(self, ghdl_version_string="GHDL 0.36 ..."):
        with patch("tsfpga.vivado.simlib_ghdl.subprocess.check_output") as check_output:
            check_output.return_value = ghdl_version_string.encode("UTF-8")

            simulator_class = MagicMock()
            simulator_class.name = "ghdl"
            simulator_class.find_prefix.return_value = "/usr/bin"

            vunit_proj = MagicMock()
            vunit_proj._simulator_class = simulator_class  # pylint: disable=protected-access

            vivado_simlib = VivadoSimlib.init(
                self.output_path, vunit_proj, Path("/tools/xilinx/Vivado/2019.2/bin/vivado"))

            return vivado_simlib

    def test_should_not_recompile(self):
        self.assert_should_compile(self.vivado_simlib)
        self.assert_should_not_compile(self.vivado_simlib)

    def test_ghdl_version_string(self):
        assert self.get_vivado_simlib(ghdl_version_string="GHDL 0.36 (v0.36) [Dunoon edition]"). \
            artifact_name.endswith(".ghdl_0_36_v0_36")
        assert self.get_vivado_simlib(ghdl_version_string="GHDL 0.36 [Dunoon edition]"). \
            artifact_name.endswith(".ghdl_0_36")
        assert self.get_vivado_simlib(ghdl_version_string="GHDL 0.36 (v0.36)"). \
            artifact_name.endswith(".ghdl_0_36_v0_36")
        assert self.get_vivado_simlib(ghdl_version_string="GHDL 0.36"). \
            artifact_name.endswith(".ghdl_0_36")

        assert self.get_vivado_simlib(ghdl_version_string="GHDL 0.37-dev (v0.36-1605-ge4aa89cd) [Dunoon edition]"). \
            artifact_name.endswith(".ghdl_0_37_dev_v0_36_1605_ge4aa89cd")
        assert self.get_vivado_simlib(ghdl_version_string="GHDL 0.37-dev (v0.36-1605-ge4aa89cd)"). \
            artifact_name.endswith(".ghdl_0_37_dev_v0_36_1605_ge4aa89cd")

    @staticmethod
    def assert_should_compile(vivado_simlib):
        assert vivado_simlib.compile_is_needed
        with patch("tsfpga.vivado.simlib_ghdl.VivadoSimlibGhdl._compile", autospec=True) as mock:
            vivado_simlib.compile_if_needed()
            mock.assert_called_once()

    @staticmethod
    def assert_should_not_compile(vivado_simlib):
        assert not vivado_simlib.compile_is_needed
        with patch("tsfpga.vivado.simlib_ghdl.VivadoSimlibGhdl._compile", autospec=True) as mock:
            vivado_simlib.compile_if_needed()
            mock.assert_not_called()



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import sys

from vunit import VUnitCLI, VUnit

from tsfpga.yosys_project import YosysProject

from tsfpga.module import BaseModule


class FormalConfig:

    # pylint: disable=too-many-arguments
    def __init__(
            self,
            top,
            generics,
            engine_command="smtbmc z3",
            mode="bmc",
            depth=20
    ):
        self.top = top
        self.generics = generics
        self.formal_settings = dict(engine_command=engine_command, mode=mode, depth=depth)

    @property
    def test_name(self):
        return BaseModule.test_case_name(self.top, self.generics)


class FormalProject:

    def __init__(self, project_path, modules):
        self.config_list = []
        self.project_path = project_path
        self.modules = modules

    def add_config(self, **kwargs):
        config = FormalConfig(**kwargs)
        self.config_list.append(config)

    def list_tests(self):
        print(80 * "=")
        print("Available tests:")
        for config in self.config_list:
            print(config.test_name)
        print(80 * "=")

    def run(self):
        # Set up a VUnit project for compilation of sources
        args = VUnitCLI().parse_args([])
        args.output_path = self.project_path / "vunit_out"
        vunit_proj = VUnit.from_args(args=args)
        vunit_proj.enable_location_preprocessing()
        vunit_proj.enable_check_preprocessing()

        for module in self.modules:
            vunit_library = vunit_proj.add_library(module.library_name)
            for hdl_file in module.get_synthesis_files():
                if hdl_file.is_vhdl:
                    vunit_library.add_source_file(hdl_file.path)
                else:
                    assert False, "Can not handle this file: " + hdl_file.path

        vunit_proj.set_compile_option("ghdl.a_flags", ["-fpsl"])
        vunit_proj._main_compile_only()  # pylint: disable=protected-access

        src_files = []
        compilation_outputs = []
        for module in self.modules:
            src_files += [src_file.path for src_file in module.get_synthesis_files()]
            compilation_outputs += [args.output_path / "ghdl" / "libraries" / module.library_name]

        result = 0
        for config in self.config_list:
            project = YosysProject(
                modules=self.modules,
                top=config.top,
                generics=config.generics,
                formal_settings=config.formal_settings
            )

            yosys_project_path = self.project_path / config.test_name
            result += project.run_formal(
                project_path=yosys_project_path,
                src_files=src_files,
                compilation_outputs=compilation_outputs)

        if result == 0:
            print("All tests passed")
        else:
            print("One or more tests failed")

        sys.exit(result)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------


class HdlFile:

    """
    Class for representing a HDL source code file.
    """

    vhdl_file_ending = ".vhd"
    verilog_source_file_ending = ".v"
    verilog_header_file_ending = ".vh"
    file_endings = (vhdl_file_ending, verilog_source_file_ending, verilog_header_file_ending)

    def __init__(self, path):
        """
        Args:
            path (`pathlib.Path`): Path to a HDL source code  file.
        """
        self.path = path

    @property
    def is_vhdl(self):
        """
        True if the file is a VHDL file. Otherwise False.
        """
        return self.path.name.endswith(self.vhdl_file_ending)

    @property
    def is_verilog_source(self):
        """
        True if the file is a Verilog source file. Otherwise False.
        """
        return self.path.name.endswith(self.verilog_source_file_ending)

    @property
    def is_verilog_header(self):
        """
        True if the file is a Verilog header file. Otherwise False.
        """
        return self.path.name.endswith(self.verilog_header_file_ending)

    def __str__(self):
        return str(self.path)



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

import copy
import unittest

import pytest

from tsfpga.system_utils import create_file
from tsfpga.register_list import from_toml, load_toml_file, RegisterList
from tsfpga.register_types import Register


def get_test_default_registers():
    registers = [
        Register("config", 0, "r_w", "Configuration register."),
    ]
    return registers


def test_deep_copy_of_registers_actually_copies_everything():
    registers = get_test_default_registers()
    for register in registers:
        if register.name == "config":
            config_register = register

    registers_copy = copy.deepcopy(registers)
    for register in registers_copy:
        if register.name == "config":
            config_register_copy = register

    config_register_copy.description = "Dummy"
    config_register_copy.bits.append("dummy object")

    assert config_register.description == "Configuration register."
    assert len(config_register.bits) == 0


def test_invalid_register_mode_should_raise_exception():
    registers = RegisterList(None, None)
    registers.append_register("test", "r_w")

    with pytest.raises(ValueError) as exception_info:
        registers.append_register("hest", "x")
    assert str(exception_info.value) == 'Invalid mode "x" for register "hest"'

    register_array = registers.append_register_array("array", 2)
    register_array.append_register("apa", "r")
    with pytest.raises(ValueError) as exception_info:
        register_array.append_register("zebra", "y")
    assert str(exception_info.value) == 'Invalid mode "y" for register "zebra"'


def test_header_constants():
    registers = RegisterList(None, None)
    registers.add_constant("test", 123)
    registers.add_constant("hest", 456)

    assert registers.constants[0].name == "test"
    assert registers.constants[0].value == 123
    assert registers.constants[1].name == "hest"
    assert registers.constants[1].value == 456


@pytest.mark.usefixtures("fixture_tmp_path")
class TestRegisterList(unittest.TestCase):

    tmp_path = None

    module_name = "sensor"
    toml_data = """\

################################################################################
[register.data]

mode = "w"
default_value = 3


################################################################################
[register.irq]

mode = "r_w"
description = "Interrupt register"

[register.irq.bits]

bad = "Bad things happen"
not_good = ""


################################################################################
[register_array.configuration]

array_length = 3

# ------------------------------------------------------------------------------
[register_array.configuration.register.input_settings]

description = "Input configuration"
mode = "r_w"
default_value = 1

[register_array.configuration.register.input_settings.bits]

enable = "Enable things"
disable = ""


# ------------------------------------------------------------------------------
[register_array.configuration.register.output_settings]

mode = "w"

[register_array.configuration.register.output_settings.bits]

enable = ""
disable = "Disable things"


################################################################################
%s
"""

    def setUp(self):
        self.toml_file = create_file(self.tmp_path / "sensor_regs.toml", self.toml_data % "")

    def test_order_of_registers_and_bits(self):
        registers = from_toml(self.module_name, self.toml_file).register_objects

        assert registers[0].name == "data"
        assert registers[0].mode == "w"
        assert registers[0].index == 0
        assert registers[0].description == ""
        assert registers[0].default_value == 3
        assert registers[0].bits == []

        assert registers[1].name == "irq"
        assert registers[1].mode == "r_w"
        assert registers[1].index == 1
        assert registers[1].description == "Interrupt register"
        assert registers[1].default_value == 0
        assert registers[1].bits[0].name == "bad"
        assert registers[1].bits[0].description == "Bad things happen"
        assert registers[1].bits[1].name == "not_good"
        assert registers[1].bits[1].description == ""

        assert registers[2].name == "configuration"
        assert registers[2].length == 3
        assert registers[2].index == 2 + 2 * 3 - 1
        assert len(registers[2].registers) == 2
        assert registers[2].registers[0].name == "input_settings"
        assert registers[2].registers[0].mode == "r_w"
        assert registers[2].registers[0].index == 0
        assert registers[2].registers[0].description == "Input configuration"
        assert registers[2].registers[0].default_value == 1
        assert registers[2].registers[0].bits[0].name == "enable"
        assert registers[2].registers[0].bits[0].description == "Enable things"
        assert registers[2].registers[0].bits[1].name == "disable"
        assert registers[2].registers[0].bits[1].description == ""
        assert registers[2].registers[1].name == "output_settings"
        assert registers[2].registers[1].mode == "w"
        assert registers[2].registers[1].index == 1
        assert registers[2].registers[1].description == ""
        assert registers[2].registers[1].default_value == 0
        assert registers[2].registers[1].bits[0].name == "enable"
        assert registers[2].registers[1].bits[0].description == ""
        assert registers[2].registers[1].bits[1].name == "disable"
        assert registers[2].registers[1].bits[1].description == "Disable things"

    def test_default_registers(self):
        default_registers = get_test_default_registers()
        num_default_registers = len(default_registers)
        toml_registers = from_toml(self.module_name, self.toml_file, default_registers)

        # The registers from this test are appended at the end
        assert toml_registers.get_register("data").index == num_default_registers
        assert toml_registers.get_register("irq").index == num_default_registers + 1

    def test_load_nonexistent_toml_file_should_raise_exception(self):
        file = self.toml_file.with_name("apa.toml")
        with pytest.raises(FileNotFoundError) as exception_info:
            load_toml_file(file)
        assert str(exception_info.value) == f"Requested TOML file does not exist: {file}"

    def test_load_dirty_toml_file_should_raise_exception(self):
        data = self.toml_data % "apa"
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            load_toml_file(self.toml_file)
        assert str(exception_info.value).startswith(f"Error while parsing TOML file {self.toml_file}:\nKey name found without value.")

    def test_plain_register_with_array_length_attribute_should_raise_exception(self):
        extras = """
[register.apa]

mode = "r_w"
array_length = 4
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value) == f"Plain register apa in {self.toml_file} can not have array_length attribute"

    def test_register_array_but_no_array_length_attribute_should_raise_exception(self):
        extras = """

[register_array.apa]

[register_array.apa.register.hest]

mode = "r_w"
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value) == f"Register array apa in {self.toml_file} does not have array_length attribute"

    def test_register_in_array_with_no_mode_attribute_should_raise_exception(self):
        extras = """
[register_array.apa]

array_length = 2

[register_array.apa.register.hest]

description = "nothing"
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value) == f"Register hest within array apa in {self.toml_file} does not have mode field"

    def test_no_mode_field_should_raise_exception(self):
        extras = """
[register.apa]

description = "w"
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value) == f"Register apa in {self.toml_file} does not have mode field"

    def test_two_registers_with_same_name_should_raise_exception(self):
        extras = """
[register.irq]

mode = "w"
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value).startswith(
            f"Error while parsing TOML file {self.toml_file}:\nWhat? irq already exists?")

    def test_register_with_same_name_as_register_array_should_raise_exception(self):
        extras = """
[register.configuration]

mode = "w"
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value) == f"Duplicate name configuration in {self.toml_file}"

    def test_two_bits_with_same_name_should_raise_exception(self):
        extras = """
[register.test_reg]

mode = "w"

[register.test_reg.bits]

test_bit = "Declaration 1"
test_bit = "Declaration 2"
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value).startswith(
            f"Error while parsing TOML file {self.toml_file}:\nDuplicate keys!")

    def test_overriding_default_register(self):
        extras = """
[register.config]

description = "apa"
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)
        toml_registers = from_toml(self.module_name, self.toml_file, get_test_default_registers())

        assert toml_registers.get_register("config").description == "apa"

    def test_changing_mode_of_default_register_should_raise_exception(self):
        extras = """
[register.config]

mode = "w"
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file, get_test_default_registers())
        assert str(exception_info.value) == f"Overloading register config in {self.toml_file}, one can not change mode from default"

    def test_unknown_register_field_should_raise_exception(self):
        extras = """
[register.test_reg]

mode = "w"
dummy = 3
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value) == f"Error while parsing register test_reg in {self.toml_file}:\nUnknown key dummy"

    def test_unknown_register_array_field_should_raise_exception(self):
        extras = """
[register_array.test_array]

array_length = 2
dummy = 3

[register_array.test_array.hest]

mode = "r"
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value) == f"Error while parsing register array test_array in {self.toml_file}:\nUnknown key dummy"

    def test_unknown_register_field_in_register_array_should_raise_exception(self):
        extras = """
[register_array.test_array]

array_length = 2

[register_array.test_array.register.hest]

mode = "r"
dummy = 3
"""
        data = self.toml_data % extras
        create_file(self.toml_file, data)

        with pytest.raises(ValueError) as exception_info:
            from_toml(self.module_name, self.toml_file)
        assert str(exception_info.value) == f"Error while parsing register hest in array test_array in {self.toml_file}:\nUnknown key dummy"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.register_code_generator import RegisterCodeGenerator
from tsfpga.register_types import Register


class RegisterCppGenerator(RegisterCodeGenerator):

    def __init__(self, module_name, generated_info):
        self.module_name = module_name
        self.generated_info = generated_info
        self._class_name = self._to_pascal_case(module_name)

    def get_interface(self, register_objects, constants):
        cpp_code = f"class I{self._class_name}\n"
        cpp_code += "{\n"
        cpp_code += "public:\n"

        for constant in constants:
            cpp_code += f"  static const int {constant.name} = {constant.value}L;\n"
        cpp_code += f"  static const size_t num_registers = {register_objects[-1].index + 1}uL;\n\n"

        for register_object in register_objects:
            if isinstance(register_object, Register):
                cpp_code += self._bit_constants(register_object)
            else:
                constant_name = self._array_length_constant_name(register_object)
                cpp_code += f"  static const size_t {constant_name} = {register_object.length}uL;\n\n"

                for register in register_object.registers:
                    cpp_code += self._bit_constants(register, register_object.name)

        cpp_code += f"  virtual ~I{self._class_name}() {{ }}\n"

        for register, register_array in self._iterate_registers(register_objects):
            cpp_code += "\n"
            if register.is_bus_readable:
                cpp_code += f"  virtual uint32_t {self._getter_function_signature(register, register_array)} const = 0;\n"
            if register.is_bus_writeable:
                cpp_code += f"  virtual void {self._setter_function_signature(register, register_array)} const = 0;\n"

        cpp_code += "};\n"

        cpp_code_top = f"""\
{self._file_header()}
#pragma once

#include <cassert>
#include <cstdint>
#include <cstdlib>

"""
        return cpp_code_top + self._with_namespace(cpp_code)

    def get_header(self, register_objects):
        cpp_code = f"class {self._class_name} : public I{self._class_name}\n"
        cpp_code += "{\n"

        cpp_code += "private:\n"
        cpp_code += "  volatile uint32_t *m_registers;\n\n"

        cpp_code += "public:\n"
        cpp_code += f"  {self._constructor_signature()};\n\n"
        cpp_code += f"  virtual ~{self._class_name}() {{ }}\n"

        for register, register_array in self._iterate_registers(register_objects):
            cpp_code += "\n"
            if register.is_bus_readable:
                cpp_code += f"  virtual uint32_t {self._getter_function_signature(register, register_array)} const override;\n"
            if register.is_bus_writeable:
                cpp_code += f"  virtual void {self._setter_function_signature(register, register_array)} const override;\n"

        cpp_code += "};\n"

        cpp_code_top = f"""\
{self._file_header()}
#pragma once

#include "i_{self.module_name}.h"

"""
        return cpp_code_top + self._with_namespace(cpp_code)

    def get_implementation(self, register_objects):
        cpp_code = f"{self._class_name}::{self._constructor_signature()}\n"
        cpp_code += "    : m_registers(reinterpret_cast<volatile uint32_t *>(base_address))\n"
        cpp_code += "{\n"
        cpp_code += "  // Empty\n"
        cpp_code += "}\n\n"

        for register, register_array in self._iterate_registers(register_objects):
            if register.is_bus_readable:
                cpp_code += self._getter_function(register, register_array)
            if register.is_bus_writeable:
                cpp_code += self._setter_function(register, register_array)

        cpp_code_top = f"{self._file_header()}\n"
        cpp_code_top += f"#include \"include/{self.module_name}.h\"\n\n"

        return cpp_code_top + self._with_namespace(cpp_code)

    @staticmethod
    def _bit_constants(register, name_prefix=None):
        cpp_code = ""
        for bit in register.bits:
            if name_prefix is None:
                bit_constant_name = f"{register.name}_{bit.name}"
            else:
                bit_constant_name = f"{name_prefix}_{register.name}_{bit.name}"

            bit_constant_value = 2 ** bit.index
            cpp_code += f"  static const uint32_t {bit_constant_name} = {bit_constant_value}uL;\n"

        if cpp_code:
            cpp_code += "\n"

        return cpp_code

    @staticmethod
    def _array_length_constant_name(register_array):
        return f"{register_array.name}_array_length"

    @staticmethod
    def _with_namespace(cpp_code_body):
        cpp_code = "namespace fpga_regs\n"
        cpp_code += "{\n\n"
        cpp_code += f"{cpp_code_body}"
        cpp_code += "\n} /* namespace fpga_regs */\n"
        return cpp_code

    @staticmethod
    def _comment(comment):
        return f"/* {comment} */\n"

    def _file_header(self):
        return self._comment(self.generated_info)

    @staticmethod
    def _to_pascal_case(snake_string):
        """
        Returns e.g., my_funny_string -> MyFunnyString
        """
        return snake_string.title().replace("_", "")

    def _constructor_signature(self):
        return f"{self._class_name}(volatile uint8_t *base_address)"

    def _setter_function(self, register, register_array):
        cpp_code = f"void {self._class_name}::{self._setter_function_signature(register, register_array)} const\n"
        cpp_code += "{\n"
        if register_array is None:
            cpp_code += f"  m_registers[{register.index}] = value;\n"
        else:
            cpp_code += f"  assert(array_index < {self._array_length_constant_name(register_array)});\n"
            cpp_code += f"  const size_t index = "\
                f"{register_array.base_index} + array_index * {len(register_array.registers)} + {register.index};\n"
            cpp_code += "  m_registers[index] = value;\n"
        cpp_code += "}\n\n"
        return cpp_code

    @staticmethod
    def _setter_function_signature(register, register_array):
        if register_array is None:
            return f"set_{register.name}(uint32_t value)"
        return f"set_{register_array.name}_{register.name}(size_t array_index, uint32_t value)"

    def _getter_function(self, register, register_array):
        cpp_code = f"uint32_t {self._class_name}::{self._getter_function_signature(register, register_array)} const\n"
        cpp_code += "{\n"
        if register_array is None:
            cpp_code += f"  return m_registers[{register.index}];\n"
        else:
            cpp_code += f"  assert(array_index < {self._array_length_constant_name(register_array)});\n"
            cpp_code += f"  const size_t index = "\
                f"{register_array.base_index} + array_index * {len(register_array.registers)} + {register.index};\n"
            cpp_code += "  return m_registers[index];\n"
        cpp_code += "}\n\n"
        return cpp_code

    @staticmethod
    def _getter_function_signature(register, register_array):
        if register_array is None:
            return f"get_{register.name}()"
        return f"get_{register_array.name}_{register.name}(size_t array_index)"



# ------------------------------------------------------------------------------
# Copyright (c) Lukas Vik. All rights reserved.
# ------------------------------------------------------------------------------

from tsfpga.register_code_generator import RegisterCodeGenerator
from tsfpga.register_types import Register


class RegisterCGenerator(RegisterCodeGenerator):

    def __init__(self, module_name, generated_info):
        self.module_name = module_name
        self.generated_info = generated_info

    def get_header(self, register_objects, constants):
        define_name = self.module_name.upper() + "_REGS_H"

        c_code = f"""\
{self._file_header()}
#ifndef {define_name}
#define {define_name}
#pragma pack(push, 1)

#include <stdint.h>

{self._register_struct(register_objects)}
{self._register_defines(register_objects)}\
{self._constants(constants)}
#pragma pack(pop)
#endif {self._comment(define_name)}"""

        return c_code

    @staticmethod
    def _comment(comment):
        return f"/* {comment} */\n"

    def _file_header(self):
        return self._comment(self.generated_info)

    def _register_struct(self, register_objects):
        array_structs = ""

        register_struct_type = f"{self.module_name}_regs_t"
        register_struct = f"typedef struct {register_struct_type}\n"
        register_struct += "{\n"
        for register_object in register_objects:
            if isinstance(register_object, Register):
                register_struct += f"  uint32_t {register_object.name};\n"
            else:
                array_struct_type = f"{self.module_name}_{register_object.name}_t"

                array_structs = f"typedef struct {array_struct_type}\n"
                array_structs += "{\n"
                for register in register_object.registers:
                    array_structs += f"  uint32_t {register.name};\n"
                array_structs += f"}} {array_struct_type};\n\n"

                register_struct += f"  {array_struct_type} {register_object.name}[{register_object.length}];\n"
        register_struct += f"}} {register_struct_type};\n"
        return array_structs + register_struct

    def _register_defines(self, register_objects):
        c_code = f"#define {self.module_name.upper()}_NUM_REGS ({register_objects[-1].index + 1}u)\n\n"

        for register, register_array in self._iterate_registers(register_objects):
            c_code += self._addr_define(register, register_array)
            c_code += self._bit_definitions(register, register_array)
            c_code += "\n"

        return c_code

    def _addr_define(self, register, register_array):
        name = self._register_define_name(register, register_array)
        if register_array is None:
            addr = register.index * 4
            return f"#define {name}_ADDR ({addr}u)\n"

        c_code = f"#define {name}_ADDR(array_index) "\
            f"(4u * ({register_array.base_index}u + (array_index) * {len(register_array.registers)}u + {register.index}u))\n"
        return c_code

    def _bit_definitions(self, register, register_array):
        register_name = self._register_define_name(register, register_array)

        c_code = ""
        for bit in register.bits:
            bit_name = f"{register_name}_{bit.name.upper()}"
            c_code += f"#define {bit_name}_BIT ({bit.index}u)\n"
            c_code += f"#define {bit_name} ({2**bit.index}u)\n"

        return c_code

    def _register_define_name(self, register, register_array):
        if register_array is None:
            name = f"{self.module_name}_{register.name}"
        else:
            name = f"{self.module_name}_{register_array.name}_{register.name}"
        return name.upper()

    def _constants(self, constants):
        c_code = ""
        for constant in constants:
            c_code += f"#define {self.module_name.upper()}_CONSTANT_{constant.name.upper()} ({constant.value})\n"
        return c_code
